/*-
 * Copyright (c) 2000 George C A Reid <greid@ukug.uk.freebsd.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer
 *    in this position and unchanged.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: plasma_saver.c,v 1.2 2001/01/24 16:56:49 greid Exp $
 * $FreeBSD$
 */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/module.h>
#include <sys/syslog.h>
#include <sys/consio.h>
#include <sys/fbio.h>

#include <dev/fb/fbreg.h>
#include <dev/fb/splashreg.h>
#include <dev/syscons/syscons.h>
#include "plasma_saver.h"

/*
 * NOTE: This code is rather heavy on resources. Run it as eye-candy
 * only
 *
 */


#define SCR_W 320
#define SCR_H 200
#define PLAS_S	128	/* Plasma is 128x128 */
#define WAVEMOD 1.3	/* Shrink/stretch factor for sine wave.
			   Effectively controls the 'speed' */

/* Miscellaneous global variables */
static u_char *video;
static int blanked;
static char xbuffer[SCR_W];
static char ybuffer[SCR_H];

static float angle[5];
static float anglebak[3];

static float
aprsin(float theta)
/*
 * An evil, evil approximation of sin(x). It computes the sin
 * value by first standardising the input into the range
 * 0 <= x <= 2pi and then returning the corresponding value from
 * the table of precomputed values in plasma_saver.h
 * This code may cause your crops to fail and demons to attack
 * your village. It is probably also responsible for the lack of
 * world peace and you will go to hell if you reuse it.
 *
 */
{
        int apos;
        int neg = 0;

        if(theta < 0) {
		neg = 1;
		theta = -theta;
        }

        if(theta > TWO_PI)
		theta -= (float)((int)(theta/TWO_PI)*TWO_PI);

	/* theta now > 0, < 2pi, look up in table */

        apos = (int)((theta/TWO_PI)*90.0);
	return((neg) ? -aprsinv[apos] : aprsinv[apos]);
}

static int
plasma_init(video_adapter_t *adp)
{
	video_info_t	info;

	/* Console capable of running 320x200x256? */
	if(get_mode_info(adp, M_VGA_CG320, &info)) {
		log(LOG_NOTICE, "plasma_saver: the console does not support M_VGA_CG320\n");
		return(ENODEV);
	}

	blanked = 0;
	return(0);
}

static int
plasma_fini(video_adapter_t *adp)
{
	return(0);
}

void draw_plasma(u_char *buffer)
{
	short i,j;
	anglebak[0] = angle[0];
	anglebak[1] = angle[1];
	anglebak[2] = angle[2];
        for (i=0; i<PLAS_S; i++)
        {
	        xbuffer[i] =	16*aprsin(angle[0]) + 8*aprsin(angle[1])
				+ 4*aprsin(angle[2]);
	        angle[0] += 2*PI/PLAS_S;
                angle[1] += 4*PI/PLAS_S;
                angle[2] += 8*PI/PLAS_S;
                angle[3] += 2*PI/PLAS_S;
                angle[4] += 4*PI/PLAS_S;
		ybuffer[i] = 16*aprsin(angle[3]) + 8*aprsin(angle[4]);
	}
	for(i=0; i<PLAS_S; i++)
	{
		/* Plot the points in the video buffer.
		   Colo(u)r is the sum of the x and y
		   buffers (generated from the 'sine'
		   functions) above and uses the last
		   128 palette entries */

                for (j=0; j<PLAS_S; j++)
			buffer[((i+45)*SCR_W) + (j+96)] = 128 + 
				ybuffer[i] + xbuffer[j];
	}
        angle[0] = anglebak[0] - ((2*PI/SCR_W)*WAVEMOD);
        angle[1] = anglebak[1] + ((6*PI/SCR_W)*WAVEMOD);
        angle[2] = anglebak[2] + ((4*PI/SCR_W)*WAVEMOD);
        angle[3] = anglebak[0] + ((6*PI/SCR_W)*WAVEMOD);
        angle[4] = anglebak[1] + ((10*PI/SCR_W)*WAVEMOD);
}

static int
plasma_saver(video_adapter_t *adp, int blank)
{
	if(blank) {
		if(blanked <= 0) {
			set_video_mode(adp, M_VGA_CG320);
	            	load_palette(adp, plasma_pal);
			blanked++;
			video = (u_char *)adp->va_window;
			/* Zero various stuff */
			memset(angle, 0, sizeof(angle));
			memset(anglebak, 0, sizeof(anglebak));
			memset(video, 0, SCR_W*SCR_H);
		}
		draw_plasma(video);
	} else {
		blanked = 0;
	}
	return(0);
}	

static scrn_saver_t plasma_module = {
	"plasma_saver", plasma_init, plasma_fini, plasma_saver, NULL
};

SAVER_MODULE(plasma_saver, plasma_module);
